"""
This module implements the WindowSensor class.
"""

import typing as T
if T.TYPE_CHECKING:
    # pylint: disable=cyclic-import,unused-import
    from .room import Room

import observable

from .. import common


class WindowSensor:
    """A sensor for Heaty's open window detection."""

    def __init__(self, entity_id: str, cfg: dict, room: "Room") -> None:
        super().__init__()
        self.entity_id = entity_id
        self.cfg = cfg
        self.room = room
        self.app = room.app
        self.events = observable.Observable()  # type: observable.Observable

    def __repr__(self) -> str:
        return "<WindowSensor {}, {}>".format(
            str(self), "open" if self.is_open() else "closed"
        )

    def __str__(self) -> str:
        return "W:{}".format(self.cfg.get("friendly_name", self.entity_id))

    def _state_cb(
            self, entity: str, attr: str,
            old: T.Optional[dict], new: T.Optional[dict],
            kwargs: dict
    ) -> None:
        """Is called when the window sensor's state has changed.
        This method triggers the opened/closed event."""

        self.log("State is now {}.".format(new),
                 level="DEBUG", prefix=common.LOG_PREFIX_INCOMING)

        self.events.trigger("open_close", self, self.is_open())

    def initialize(self) -> None:
        """Should be called in order to register state listeners and
        timers."""

        self.log("Initializing window sensor (entity_id={})."
                 .format(repr(self.entity_id)),
                 level="DEBUG")

        self.log("Listening for state changes (delay={})."
                 .format(self.cfg["delay"]),
                 level="DEBUG")
        self.app.listen_state(self._state_cb, self.entity_id,
                              duration=self.cfg["delay"])

    def is_open(self) -> bool:
        """Returns whether the sensor reports open or not."""

        open_state = self.cfg["open_state"]
        states = []
        if isinstance(open_state, list):
            states.extend(open_state)
        else:
            states.append(open_state)
        return self.app.get_state(self.entity_id) in states

    def log(self, msg: str, *args: T.Any, **kwargs: T.Any) -> None:
        """Prefixes the window sensor to log messages."""
        msg = "[{}] {}".format(self, msg)
        self.room.log(msg, *args, **kwargs)



"""
This module implements the StatisticsZone class for collecting statistics.
"""

import typing as T
if T.TYPE_CHECKING:
    # pylint: disable=cyclic-import,unused-import
    import uuid
    # pylint: disable=cyclic-import,unused-import
    from .app import HeatyApp
    from .room import Room

from .. import common
from . import util


class _WeightedValue:
    """A measured value having a weight for proper average calculation."""

    def __init__(self, value: float, weight: float) -> None:
        self.value = value
        self.weight = weight

    def __repr__(self) -> str:
        return "{}(weight={})" \
               .format(util.format_sensor_value(self.value), self.weight)


class StatisticsZone:
    """A zone (group of rooms) used for collecting statistical data."""

    def __init__(self, name: str, cfg: T.Dict, app: "HeatyApp") -> None:
        self.name = name
        self.cfg = cfg
        self.app = app
        self.rooms = []  # type: T.List[Room]
        self._stats_timer = None  # type: T.Optional[uuid.UUID]

    def __repr__(self) -> str:
        return "<StatisticsZone {}>".format(self.name)

    def __str__(self) -> str:
        return "SZ:{}".format(self.cfg.get("friendly_name", self.name))

    def _collect_temp_delta(self, cfg: dict) -> T.List[_WeightedValue]:
        off_value = cfg["off_value"]
        values = []
        for room in self.rooms:
            for therm in room.thermostats:
                weight = cfg["thermostat_weights"].get(therm.entity_id, 1)
                if weight == 0:
                    # ignore this thermostat
                    continue

                if therm.current_temp is None or \
                   therm.current_target_temp is None or \
                   therm.current_temp.is_off() or \
                   therm.current_target_temp.is_off():
                    if off_value is None:
                        # thermostats that are off should be excluded
                        continue
                    temp_delta = float(off_value)
                else:
                    temp_delta = float(therm.current_target_temp -
                                       therm.current_temp)
                    factor = cfg["thermostat_factors"].get(therm.entity_id, 1)
                    temp_delta *= factor
                value = _WeightedValue(temp_delta, weight)
                self.log("Value for {} in {} is {}".format(therm, room, value),
                         level="DEBUG")
                values.append(value)
        return values

    def _do_update_stats(self) -> None:
        """Writes the zone statistics to Home Assistant."""

        self._stats_timer = None

        if not self.cfg["parameters"]:
            self.log("No parameters configured, nothing to update.",
                     level="DEBUG")
            return

        params = {}  # type: T.Dict[str, T.List[_WeightedValue]]
        for param in self.cfg["parameters"]:
            self.log("Collecting {}".format(param),
                     level="DEBUG")
            cfg = self.cfg["parameters"][param]
            params[param] = getattr(self, "_collect_{}".format(param))(cfg)

        fmt = util.format_sensor_value
        for param, values in params.items():
            _min = fmt(min([v.value for v in values]) if values else 0)
            _avg = fmt(sum([v.value * v.weight for v in values]) /
                       sum([v.weight for v in values]) if values else 0)
            _max = fmt(max([v.value for v in values]) if values else 0)
            self.log("{} (min/avg/max): {} / {} / {}"
                     .format(param, _min, _avg, _max),
                     level="DEBUG")
            self._set_sensor("min_{}".format(param), _min)
            self._set_sensor("avg_{}".format(param), _avg)
            self._set_sensor("max_{}".format(param), _max)

    def _set_sensor(self, param: str, state: T.Any) -> None:
        """Updates the sensor for given parameter in HA."""

        entity_id = "sensor.heaty_{}_zone_{}_{}" \
                    .format(self.app.cfg["heaty_id"], self.name, param)
        self.log("Setting state of {} to {}."
                 .format(repr(entity_id), repr(state)),
                 level="DEBUG", prefix=common.LOG_PREFIX_OUTGOING)
        self.app.set_state(entity_id, state=state)

    def initialize(self) -> None:
        """Fetches the Room objects and sets up internal data structures
        before triggering an initial statistics update."""

        self.log("Initializing statistics zone (name={})."
                 .format(repr(self.name)),
                 level="DEBUG")

        if not self.cfg["parameters"]:
            self.log("No parameters configured.", level="WARNING")

        for room_name in self.cfg["rooms"]:
            room = self.app.get_room(room_name)
            if room is None:
                self.log("Room named '{}' not found, not adding it to "
                         "statistics zone."
                         .format(room_name),
                         level="ERROR")
                continue
            self.rooms.append(room)
            for therm in room.thermostats:
                self.log("Listening for temperature changes of {} in {}."
                         .format(therm, room),
                         level="DEBUG")
                therm.events.on("current_temp_changed",
                                lambda *a, **kw: self.update_stats())
                therm.events.on("target_temp_changed",
                                lambda *a, **kw: self.update_stats())

        if not self.rooms:
            self.log("No rooms configured.", level="WARNING")

        self.update_stats()

    def log(self, msg: str, *args: T.Any, **kwargs: T.Any) -> None:
        """Prefixes the zone to log messages."""
        msg = "[{}] {}".format(self, msg)
        self.app.log(msg, *args, **kwargs)

    def update_stats(self) -> None:
        """Registers a timer for sending statistics to HA in 3 seconds."""

        if self._stats_timer:
            self.log("Statistics update  pending already.",
                     level="DEBUG")
            return

        self.log("Going to update statistics in 3 seconds.",
                 level="DEBUG")
        self._stats_timer = self.app.run_in(
            lambda *a: self._do_update_stats(), 3
        )



"""
This module implements the Thermostat class.
"""

import typing as T
if T.TYPE_CHECKING:
    # pylint: disable=cyclic-import,unused-import
    import uuid
    from .room import Room

import copy
import observable

from .. import common
from . import expr


class Thermostat:
    """A thermostat to be controlled by Heaty."""

    # pylint: disable=too-many-instance-attributes

    def __init__(self, entity_id: str, cfg: dict, room: "Room") -> None:
        self.entity_id = entity_id
        self.cfg = cfg
        self.room = room
        self.app = room.app
        self.current_temp = None  # type: T.Optional[expr.Temp]
        self.current_target_temp = None  # type: T.Optional[expr.Temp]
        self.wanted_temp = None  # type: T.Optional[expr.Temp]
        self.resend_timer = None  # type: T.Optional[uuid.UUID]
        self.events = observable.Observable()  # type: observable.Observable

    def __repr__(self) -> str:
        return "<Thermostat {}>".format(str(self))

    def __str__(self) -> str:
        return "T:{}".format(self.cfg.get("friendly_name", self.entity_id))

    def _check_config_plausibility(self) -> None:
        """Is called during initialization to warn the user about some
        possible common configuration mistakes."""

        _state = self.app.get_state(self.entity_id, attribute="all")
        if not _state:
            self.log("Thermostat couldn't be found.", level="WARNING")
            return

        state = copy.deepcopy((_state or {}).get("attributes", {}))
        state.update(copy.deepcopy(_state or {}))

        required_attrs = []
        if self.cfg["supports_opmodes"]:
            required_attrs.append(self.cfg["opmode_state_attr"])
        if self.cfg["supports_temps"]:
            required_attrs.append(self.cfg["target_temp_state_attr"])
        if not required_attrs:
            self.log("At least one of supports_opmodes and "
                     "supports_temps should be enabled. "
                     "Please check your config!",
                     level="WARNING")
        for attr in required_attrs:
            if attr not in state:
                self.log("Thermostat has no attribute named {}. "
                         "Available attributes are {}. "
                         "Please check your config!"
                         .format(repr(attr), list(state.keys())),
                         level="WARNING")

        if self.cfg["supports_temps"]:
            value = state.get(self.cfg["target_temp_state_attr"])
            try:
                value = float(value)
            except (TypeError, ValueError):
                self.log("The value {} for attribute {} is no valid "
                         "temperature value. "
                         "Please check your config!"
                         .format(repr(value),
                                 self.cfg["target_temp_state_attr"]),
                         level="WARNING")

        allowed_opmodes = state.get("operation_list")
        if not self.cfg["supports_opmodes"]:
            if allowed_opmodes:
                self.log("Operation mode support has been disabled, "
                         "but the following modes seem to be supported: {} "
                         "Maybe disabling it was a mistake?"
                         .format(allowed_opmodes),
                         level="WARNING")
            return

        if self.cfg["opmode_state_attr"] != "operation_mode":
            # we can't rely on operation_list in this case
            return
        if not allowed_opmodes:
            self.log("Attributes for thermostat contain no "
                     "'operation_list', Consider disabling "
                     "operation mode support.",
                     level="WARNING")
            return
        for opmode in (self.cfg["opmode_heat"], self.cfg["opmode_off"]):
            if opmode not in allowed_opmodes:
                self.log("Thermostat doesn't seem to support the "
                         "operation mode {}, supported modes are: {}. "
                         "Please check your config!"
                         .format(opmode, allowed_opmodes),
                         level="WARNING")

    def _state_cb(
            self, entity: str, attr: str,
            old: T.Optional[dict], new: T.Optional[dict],
            kwargs: dict, no_reschedule: bool = False
    ) -> None:
        """Is called when the thermostat's state changes.
        This method fetches both the current and target temperature from
        the thermostat and reacts accordingly."""

        attrs = copy.deepcopy((new or {}).get("attributes", {}))
        attrs.update(copy.deepcopy(new or {}))

        _target_temp = None  # type: T.Union[None, expr.Off, float, int]
        if self.cfg["supports_opmodes"]:
            opmode = attrs.get(self.cfg["opmode_state_attr"])
            self.log("Attribute {} is {}."
                     .format(self.cfg["opmode_state_attr"], opmode),
                     level="DEBUG", prefix=common.LOG_PREFIX_INCOMING)
            if opmode is None:
                # don't consider this thermostat
                return
            elif opmode == self.cfg["opmode_off"]:
                _target_temp = expr.Off()
        else:
            opmode = None

        if _target_temp is None:
            if self.cfg["supports_temps"]:
                _target_temp = attrs.get(self.cfg["target_temp_state_attr"])
                self.log("Attribute {} is {}."
                         .format(self.cfg["target_temp_state_attr"],
                                 _target_temp),
                         level="DEBUG", prefix=common.LOG_PREFIX_INCOMING)
            else:
                _target_temp = 0

        try:
            target_temp = expr.Temp(_target_temp)
        except ValueError:
            # no valid temperature, don't consider this thermostat
            return

        _current_temp = attrs.get(self.cfg["current_temp_state_attr"])
        self.log("Attribute {} is {}."
                 .format(self.cfg["current_temp_state_attr"], _current_temp),
                 level="DEBUG", prefix=common.LOG_PREFIX_INCOMING)
        try:
            current_temp = expr.Temp(_current_temp)  # type: T.Optional[expr.Temp]
        except ValueError:
            current_temp = None
        if current_temp != self.current_temp:
            self.current_temp = current_temp
            self.events.trigger("current_temp_changed", self, current_temp)

        if target_temp == self.wanted_temp:
            # thermostat adapted to the temperature we want,
            # cancel any re-send timer
            self.cancel_resend_timer()

        if target_temp != self.current_target_temp:
            self.current_target_temp = target_temp

            if self.cfg["supports_temps"]:
                self.log("Received target temperature of {}."
                         .format(str(target_temp)),
                         prefix=common.LOG_PREFIX_INCOMING)
                target_temp -= self.cfg["delta"]
            else:
                self.log("Received state of {}."
                         .format("OFF" if target_temp.is_off() else "ON"),
                         prefix=common.LOG_PREFIX_INCOMING)

            self.events.trigger(
                "target_temp_changed", self, target_temp,
                no_reschedule=no_reschedule
            )

    def initialize(self) -> None:
        """Should be called in order to register state listeners and
        timers."""

        self.log("Initializing thermostat (entity_id={})."
                 .format(repr(self.entity_id)),
                 level="DEBUG")

        self._check_config_plausibility()

        # only consider one thermostat per room
        if self.room.wanted_temp is None:
            self.log("Getting current state from thermostat.",
                     level="DEBUG")
            state = self.app.get_state(self.entity_id, attribute="all")
            if state is None:
                # unknown entity
                self.log("State for thermostat is None, ignoring it.",
                         level="WARNING")
            else:
                # populate self.current_target_temp etc. by simulating a
                # state change
                self._state_cb(self.entity_id, "all", state, state, {},
                               no_reschedule=True)

        self.log("Listening for state changes.",
                 level="DEBUG")
        self.app.listen_state(self._state_cb, self.entity_id, attribute="all")

    def log(self, msg: str, *args: T.Any, **kwargs: T.Any) -> None:
        """Prefixes the thermostat to log messages."""
        msg = "[{}] {}".format(self, msg)
        self.room.log(msg, *args, **kwargs)

    def cancel_resend_timer(self) -> bool:
        """Cancel the resend timer for this thermostat, if one exists.
        Returns whether a timer was cancelled."""

        timer = self.resend_timer
        if timer is None:
            return False
        self.app.cancel_timer(timer)
        self.resend_timer = None
        self.log("Cancelled resend timer.", level="DEBUG")
        return True

    def is_synced(self) -> bool:
        """Returns whether the thermostat's target temperature is the
        wanted temperature and no re-sending is in progress."""

        return self.resend_timer is None and \
               self.current_target_temp is not None and \
               self.current_target_temp == self.wanted_temp

    def set_temp(
            self, target_temp: expr.Temp, force_resend: bool = False
    ) -> T.Optional[expr.Temp]:
        """Sets the given target temperature on this thermostat
        The temperature should be the desired room temperature,
        without delta added. This method will try best to achieve
        the closest possible temperature supported by this particular
        thermostat.
        A temperature won't be send to the thermostat redundantly
        unless force_resend is True.
        The return value is either the actually set temperature or
        None, if nothing has been sent."""

        if target_temp.is_off():
            temp = None
            opmode = self.cfg["opmode_off"]
        else:
            temp = target_temp + self.cfg["delta"]
            if isinstance(self.cfg["min_temp"], expr.Temp) and \
               temp < self.cfg["min_temp"]:
                temp = None
                opmode = self.cfg["opmode_off"]
            else:
                opmode = self.cfg["opmode_heat"]
                if isinstance(self.cfg["max_temp"], expr.Temp) and \
                   temp > self.cfg["max_temp"]:
                    temp = self.cfg["max_temp"]

        if not self.cfg["supports_opmodes"]:
            if opmode == self.cfg["opmode_off"]:
                self.log("Not turning off because it doesn't support "
                         "operation modes.",
                         level="DEBUG")
                if self.cfg["min_temp"] is not None:
                    self.log("Setting to minimum supported temperature "
                             "instead.",
                             level="DEBUG")
                    temp = self.cfg["min_temp"]
            opmode = None

        if not self.cfg["supports_temps"] and temp is not None:
            self.log("Not setting temperature because thermostat doesn't "
                     "support temperatures.",
                     level="DEBUG")
            temp = None

        if opmode is None and temp is None:
            self.log("Nothing to send to this thermostat.",
                     level="DEBUG")
            return None

        if opmode == self.cfg["opmode_off"]:
            wanted_temp = expr.Temp(expr.OFF)  # type: T.Optional[expr.Temp]
        elif self.cfg["supports_temps"]:
            wanted_temp = temp
        else:
            wanted_temp = expr.Temp(0)
        self.wanted_temp = wanted_temp

        if not force_resend and self.is_synced():
            self.log("Not sending temperature redundantly.",
                     level="DEBUG")
            return None

        left_retries = self.cfg["set_temp_retries"]
        self.cancel_resend_timer()
        timer = self.app.run_in(
            self._set_temp_resend_cb, 1,
            left_retries=left_retries,
            opmode=opmode, temp=temp
        )
        self.resend_timer = timer

        return wanted_temp

    def _set_temp_resend_cb(self, kwargs: dict) -> None:
        """This callback sends the operation_mode and temperature to the
        thermostat. Expected values for kwargs are:
        - opmode and temp (incl. delta)
        - left_retries (after this round)"""

        opmode = kwargs["opmode"]
        temp = kwargs["temp"]
        left_retries = kwargs["left_retries"]

        self.cancel_resend_timer()

        self.log("Setting temperature = {}, operation mode = {}, "
                 "left retries = {}"
                 .format(temp or "<unset>", opmode or "<unset>",
                         left_retries),
                 level="DEBUG", prefix=common.LOG_PREFIX_OUTGOING)

        if opmode is not None:
            if opmode == self.cfg["opmode_heat"]:
                opmode_service = self.cfg["opmode_heat_service"]
                opmode_service_attr = self.cfg["opmode_heat_service_attr"]
            else:
                opmode_service = self.cfg["opmode_off_service"]
                opmode_service_attr = self.cfg["opmode_off_service_attr"]
            attrs = {"entity_id": self.entity_id}
            if opmode_service_attr:
                attrs[opmode_service_attr] = opmode
            self.app.call_service(opmode_service, **attrs)
        if temp is not None:
            attrs = {"entity_id": self.entity_id,
                     self.cfg["target_temp_service_attr"]: temp.value}
            self.app.call_service(self.cfg["target_temp_service"], **attrs)

        if not left_retries:
            return

        interval = self.cfg["set_temp_retry_interval"]
        self.log("Re-sending in {} seconds."
                 .format(interval),
                 level="DEBUG")
        timer = self.app.run_in(
            self._set_temp_resend_cb, interval,
            left_retries=left_retries - 1,
            opmode=opmode, temp=temp)
        self.resend_timer = timer



"""
This module contains the CONFIG_SCHEMA for validation with voluptuous.
"""

import typing as T

import voluptuous as vol

from . import expr, schedule, util
from .room import Room
from .thermostat import Thermostat
from .window_sensor import WindowSensor
from .stats import StatisticsZone


def build_schedule_rule(rule: dict) -> schedule.Rule:
    """Builds and returns a schedule rule from the given rule
    definition."""

    constraints = {}
    for name, value in rule.items():
        if name in schedule.Rule.CONSTRAINTS:
            constraints[name] = value

    kwargs = {
        "start_time": rule["start"],
        "end_time": rule["end"],
        "end_plus_days": rule["end_plus_days"],
        "constraints": constraints,
        "temp_expr": rule.get("temp"),
    }

    if "rules" in rule:
        return schedule.SubScheduleRule(rule["rules"], **kwargs)
    return schedule.Rule(**kwargs)

def build_schedule(rules: T.Iterable[dict]) -> schedule.Schedule:
    """Compiles the given rules and returns a schedule containing them.
    When check is True, a check is run after schedule creation to ensure
    each path contains at least one rule with a temperature expression.
    A ValueError is raised when this check fails."""

    sched = schedule.Schedule()
    for rule in rules:
        sched.rules.append(build_schedule_rule(rule))
    return sched

def config_post_hook(cfg: dict) -> dict:
    """Creates Room and other objects after config has been parsed."""

    # pylint: disable=too-many-locals

    # Build room objects.
    rooms = []
    for room_name, room_data in cfg["rooms"].items():
        therms = {}
        wsensors = {}

        # copy settings from defaults sections to this room
        for therm_name, therm_data in room_data["thermostats"].items():
            for key, val in cfg["thermostat_defaults"].items():
                therm_data.setdefault(key, val)
            therm_data = THERMOSTAT_SCHEMA(therm_data)
            therms[therm_name] = therm_data
        for wsensor_name, wsensor_data in room_data["window_sensors"].items():
            for key, val in cfg["window_sensor_defaults"].items():
                wsensor_data.setdefault(key, val)
            wsensor_data = WINDOW_SENSOR_SCHEMA(wsensor_data)
            wsensors[wsensor_name] = wsensor_data

        # complete the room's schedule.
        sched = cfg["schedule_prepend"] + room_data["schedule"] + \
                cfg["schedule_append"]

        del room_data["thermostats"]
        del room_data["window_sensors"]
        del room_data["schedule"]

        room = Room(room_name, room_data, cfg["_app"])
        rooms.append(room)

        # Create thermostat and window sensor objects and attach to room.
        for therm_name, therm_data in therms.items():
            therm = Thermostat(therm_name, therm_data, room)
            room.thermostats.append(therm)
        for wsensor_name, wsensor_data in wsensors.items():
            wsensor = WindowSensor(wsensor_name, wsensor_data, room)
            room.window_sensors.append(wsensor)

        room.schedule = sched
    del cfg["rooms"], cfg["schedule_prepend"], cfg["schedule_append"]
    cfg["_app"].rooms = rooms

    szones = []
    for zone_name, zone_cfg in cfg["statistics"].items():
        zone = StatisticsZone(zone_name, zone_cfg, cfg["_app"])
        szones.append(zone)
    del cfg["statistics"]
    cfg["_app"].stats_zones = szones

    return cfg

def validate_rule_paths(sched: schedule.Schedule) -> schedule.Schedule:
    """A validator to be run after schedule creation to ensure
    each path contains at least one rule with a temperature expression.
    A ValueError is raised when this check fails."""

    for path in sched.unfold():
        schedule.get_rule_path_temp(path)

    return sched


########## MISCELLANEOUS

ENTITY_ID_SCHEMA = vol.Schema(vol.Match(r"^[A-Za-z_]+\.[A-Za-z0-9_]+$"))
PYTHON_VAR_SCHEMA = vol.Schema(vol.Match(r"^[a-zA-Z_]+[a-zA-Z0-9_]*$"))
RANGE_STRING_SCHEMA = vol.Schema(vol.All(
    vol.Any(
        int,
        vol.Match(r"^ *\d+( *\- *\d+)?( *\, *\d+( *\- *\d+)?)* *$"),
    ),
    util.expand_range_string,
))
PARTIAL_DATE_SCHEMA = vol.Schema({
    vol.Optional("year"): vol.All(int, vol.Range(min=1970, max=9999)),
    vol.Optional("month"): vol.All(int, vol.Range(min=1, max=12)),
    vol.Optional("day"): vol.All(int, vol.Range(min=1, max=31)),
})
TIME_SCHEMA = vol.Schema(vol.All(
    vol.Match(util.TIME_REGEXP),
    util.parse_time_string,
))
TEMP_SCHEMA = vol.Schema(vol.All(
    vol.Any(float, int, expr.Off, vol.All(str, lambda v: v.upper(), "OFF")),
    lambda v: expr.Temp(v),  # pylint: disable=unnecessary-lambda
))
TEMP_EXPRESSION_SCHEMA = vol.Schema(str)

# This schema does no real validation and default value insertion,
# it just ensures a dictionary containing dictionaries is returned.
DICTS_IN_DICT_SCHEMA = vol.Schema(vol.All(
    lambda v: v or {},
    {vol.Extra: lambda v: v or {}},
))

TEMP_EXPRESSION_MODULE_SCHEMA = vol.Schema(vol.All(
    lambda v: v or {},
    {
        "as": PYTHON_VAR_SCHEMA,
    },
))
TEMP_EXPRESSION_MODULES_SCHEMA = vol.Schema({
    vol.Extra: lambda v: TEMP_EXPRESSION_MODULE_SCHEMA(v or {}),
})


########## THERMOSTATS

THERMOSTAT_SCHEMA = vol.Schema(vol.All(
    lambda v: v or {},
    {
        "friendly_name": str,
        vol.Optional("delta", default=0): vol.Any(float, int),
        vol.Optional("min_temp", default=None): vol.Any(
            vol.All(TEMP_SCHEMA, vol.NotIn([expr.Temp(expr.OFF)])),
            None,
        ),
        vol.Optional("max_temp", default=None): vol.Any(
            vol.All(TEMP_SCHEMA, vol.NotIn([expr.Temp(expr.OFF)])),
            None,
        ),
        vol.Optional("set_temp_retries", default=10):
            vol.All(int, vol.Range(min=-1)),
        vol.Optional("set_temp_retry_interval", default=30):
            vol.All(int, vol.Range(min=1)),
        vol.Optional("supports_opmodes", default=True): bool,
        vol.Optional("supports_temps", default=True): bool,
        vol.Optional("opmode_heat", default="Heat"): str,
        vol.Optional("opmode_off", default="Off"): str,
        vol.Optional(
            "opmode_heat_service", default="climate/set_operation_mode"
        ): str,
        vol.Optional(
            "opmode_off_service", default="climate/set_operation_mode"
        ): str,
        vol.Optional("opmode_heat_service_attr", default="operation_mode"):
            vol.Any(str, None),
        vol.Optional("opmode_off_service_attr", default="operation_mode"):
            vol.Any(str, None),
        vol.Optional("opmode_state_attr", default="operation_mode"): str,
        vol.Optional(
            "target_temp_service", default="climate/set_temperature"
        ): str,
        vol.Optional("target_temp_service_attr", default="temperature"): str,
        vol.Optional("target_temp_state_attr", default="temperature"): str,
        vol.Optional(
            "current_temp_state_attr", default="current_temperature"
        ): str,
    },
))


########## WINDOW SENSORS

STATE_SCHEMA = vol.Schema(vol.Any(float, int, str))
WINDOW_SENSOR_SCHEMA = vol.Schema(vol.All(
    lambda v: v or {},
    {
        "friendly_name": str,
        vol.Optional("delay", default=10): vol.All(int, vol.Range(min=0)),
        vol.Optional("open_state", default="on"):
            vol.Any(STATE_SCHEMA, [STATE_SCHEMA]),
    },
))


########## SCHEDULES

SCHEDULE_RULE_SCHEMA = vol.Schema(vol.All(
    lambda v: v or {},
    {
        "rules": lambda v: SCHEDULE_SCHEMA(v),  # type: ignore  # pylint: disable=unnecessary-lambda
        "temp": vol.Any(TEMP_SCHEMA, TEMP_EXPRESSION_SCHEMA),
        vol.Optional("start", default=None): vol.Any(TIME_SCHEMA, None),
        vol.Optional("end", default=None): vol.Any(TIME_SCHEMA, None),
        vol.Optional("end_plus_days", default=0):
            vol.All(int, vol.Range(min=0)),
        vol.Optional("years"): RANGE_STRING_SCHEMA,
        vol.Optional("months"): RANGE_STRING_SCHEMA,
        vol.Optional("days"): RANGE_STRING_SCHEMA,
        vol.Optional("weeks"): RANGE_STRING_SCHEMA,
        vol.Optional("weekdays"): RANGE_STRING_SCHEMA,
        vol.Optional("start_date"): PARTIAL_DATE_SCHEMA,
        vol.Optional("end_date"): PARTIAL_DATE_SCHEMA,
    },
))
SCHEDULE_SCHEMA = vol.Schema(vol.All(
    lambda v: v or [],
    [SCHEDULE_RULE_SCHEMA],
    build_schedule,
))

SCHEDULE_SNIPPETS_SCHEMA = vol.Schema(vol.All(
    lambda v: v or {},
    {
        vol.Extra: vol.All(
            SCHEDULE_SCHEMA,
            validate_rule_paths,
        ),
    },
))


########## ROOMS

ROOM_SCHEMA = vol.Schema(vol.All(
    lambda v: v or {},
    {
        "friendly_name": str,
        vol.Optional("replicate_changes", default=True): bool,
        vol.Optional("reschedule_delay", default=0):
            vol.All(int, vol.Range(min=0)),
        vol.Optional("thermostats", default=dict): DICTS_IN_DICT_SCHEMA,
        vol.Optional("window_sensors", default=dict): DICTS_IN_DICT_SCHEMA,
        vol.Optional("schedule", default=list): vol.All(
            SCHEDULE_SCHEMA,
            validate_rule_paths,
        ),
    },
))


########## STATISTICS

STATS_ZONE_ROOM_SCHEMA = vol.Schema(vol.All(
    lambda v: v or {},
    {
        # More parameters may be added here in the future.
    },
))

STATS_ZONE_PARAM_THERMOSTAT_SETTINGS_ADDIN = {
    vol.Optional("thermostat_factors", default=dict): vol.All(
        lambda v: v or {},
        {
            vol.Extra: vol.All(vol.Any(float, int),
                               vol.Range(min=0), min_included=False),
        },
    ),
    vol.Optional("thermostat_weights", default=dict): vol.All(
        lambda v: v or {},
        {
            vol.Extra: vol.All(vol.Any(float, int), vol.Range(min=0)),
        },
    ),
}

STATS_ZONE_SCHEMA = vol.Schema(vol.All(
    lambda v: v or {},
    {
        "friendly_name": str,
        vol.Optional("rooms", default=dict): vol.All(
            lambda v: v or {},
            {vol.Extra: STATS_ZONE_ROOM_SCHEMA},
        ),
        vol.Optional("parameters", default=dict): vol.All(
            lambda v: v or {},
            {
                "temp_delta": vol.All(
                    lambda v: v or {},
                    util.mixin_dict({
                        vol.Optional("off_value", default=0): vol.Any(
                            float, int, None
                        ),
                    }, STATS_ZONE_PARAM_THERMOSTAT_SETTINGS_ADDIN),
                ),
            },
        ),
    },
))


########## MAIN CONFIG SCHEMA

CONFIG_SCHEMA = vol.Schema(vol.All(
    vol.Schema({
        vol.Optional("heaty_id", default="default"): str,
        vol.Optional("untrusted_temp_expressions", default=False): bool,
        vol.Optional("master_switch", default=None):
            vol.Any(ENTITY_ID_SCHEMA, None),
        vol.Optional("master_off_temp", default=expr.OFF): TEMP_SCHEMA,
        vol.Optional("window_open_temp", default=expr.OFF): TEMP_SCHEMA,
        vol.Optional("temp_expression_modules", default=dict):
            TEMP_EXPRESSION_MODULES_SCHEMA,
        vol.Optional("thermostat_defaults", default=dict):
            lambda v: v or {},
        vol.Optional("window_sensor_defaults", default=dict):
            lambda v: v or {},
        vol.Optional("schedule_prepend", default=list): vol.All(
            SCHEDULE_SCHEMA,
            validate_rule_paths,
        ),
        vol.Optional("schedule_append", default=list): vol.All(
            SCHEDULE_SCHEMA,
            validate_rule_paths,
        ),
        vol.Optional("schedule_snippets", default=dict):
            SCHEDULE_SNIPPETS_SCHEMA,
        vol.Optional("rooms", default=dict): vol.All(
            lambda v: v or {},
            {vol.Extra: ROOM_SCHEMA},
        ),
        vol.Optional("statistics", default=dict): vol.All(
            lambda v: v or {},
            {vol.Extra: STATS_ZONE_SCHEMA},
        ),
    }, extra=True),
    config_post_hook,
))



"""
This module implements the Schedule and Rule classes.
"""

import typing as T  # pylint: disable=unused-import

import collections
import datetime

from . import expr, util


# Type of a rule path, a tuple of Rule objects representing  the hirarchy
# of SubScheduleRules that lead to a Rule.
RULE_PATH_TYPE = T.Tuple["Rule", ...]


class Rule:
    """A rule that can be added to a schedule."""

    # names of schedule rule constraints to be fetched from a rule definition
    CONSTRAINTS = ("years", "months", "days", "weeks", "weekdays",
                   "start_date", "end_date")

    def __init__(
            self,
            start_time: datetime.time = None, end_time: datetime.time = None,
            end_plus_days: int = 0, constraints: T.Dict[str, T.Any] = None,
            temp_expr: expr.EXPR_TYPE = None
        ) -> None:

        if start_time is None:
            # make it midnight
            start_time = datetime.time(0, 0)
        self.start_time = start_time

        if end_time is None:
            # make it midnight (00:00 of the next day)
            end_time = datetime.time(0, 0)
            end_plus_days += 1
        self.end_time = end_time

        if end_time <= start_time and end_plus_days == 0:
            end_plus_days = 1
        self.end_plus_days = end_plus_days

        if constraints is None:
            constraints = {}
        self.constraints = constraints

        self.temp_expr = None  # type: T.Optional[expr.EXPR_TYPE]
        self.temp_expr_raw = None  # type: T.Optional[expr.EXPR_TYPE]
        if temp_expr is not None:
            if isinstance(temp_expr, str):
                temp_expr = temp_expr.strip()
            self.temp_expr_raw = temp_expr
            try:
                temp = expr.Temp(temp_expr)
            except ValueError:
                # this is a temperature expression, precompile it
                self.temp_expr = compile(temp_expr, "temp_expr", "eval")  # type: expr.EXPR_TYPE
            else:
                self.temp_expr = temp

    def __repr__(self) -> str:
        return "<Rule {}>".format(
            ", ".join(["{}={}".format(k, v)
                       for k, v in self._get_repr_properties().items()])
        )

    def _get_repr_properties(self) -> T.Dict[str, T.Any]:
        """Returns an OrderedDict with properties to be shown in repr()."""

        props = collections.OrderedDict()  # type: T.Dict[str, T.Any]
        if self.is_always_valid():
            props["always_valid"] = "yes"
        else:
            props["start"] = self.start_time
            props["end"] = self.end_time
            if self.end_plus_days != 0:
                props["end_plus_days"] = self.end_plus_days
            if self.constraints:
                props["constraints"] = list(self.constraints)
        props["temp"] = self.temp_expr_raw
        return props

    def check_constraints(self, date: datetime.date) -> bool:
        """Checks all constraints of this rule against the given date
        and returns whether they are fulfilled"""

        # pylint: disable=too-many-return-statements

        year, week, weekday = date.isocalendar()
        for constraint, allowed in self.constraints.items():
            if constraint == "years" and year not in allowed:
                return False
            if constraint == "months" and date.month not in allowed:
                return False
            if constraint == "days" and date.day not in allowed:
                return False
            if constraint == "weeks" and week not in allowed:
                return False
            if constraint == "weekdays" and weekday not in allowed:
                return False
            if constraint == "start_date" and \
               date < util.build_date_from_constraint(allowed, date, 1):
                return False
            if constraint == "end_date" and \
               date > util.build_date_from_constraint(allowed, date, -1):
                return False
        return True

    def is_always_valid(self) -> bool:
        """Returns whether this rule is universally valid (has no
        constraints and duration >= 1 day)."""

        if self.constraints:
            return False
        if self.end_time < self.start_time:
            return self.end_plus_days > 1
        return self.end_plus_days >= 1


class SubScheduleRule(Rule):
    """A schedule rule with a sub-schedule attached."""

    def __init__(
            self, sub_schedule: "Schedule",
            *args: T.Any, **kwargs: T.Any
        ) -> None:

        super().__init__(*args, **kwargs)

        self.sub_schedule = sub_schedule

    def _get_repr_properties(self) -> T.Dict[str, T.Any]:
        """Adds the sub-schedule information to repr()."""

        props = super()._get_repr_properties()
        props["sub_schedule"] = "yes"
        return props


class Schedule:
    """Holds the schedule for a room with all its rules."""

    def __init__(self, rules: T.Iterable[Rule] = None) -> None:
        self.rules = []  # type: T.List[Rule]
        if rules is not None:
            self.rules.extend(rules)

    def __add__(self, other: "Schedule") -> "Schedule":
        if not isinstance(other, type(self)):
            raise ValueError("{} objects may not be added to {}."
                             .format(type(other), self))
        return Schedule(self.rules + other.rules)

    def __repr__(self) -> str:
        return "<Schedule with {} rules>".format(len(self.rules))

    def matching_rules(
            self, when: datetime.datetime
        ) -> T.Iterator[RULE_PATH_TYPE]:
        """Returns an iterator over paths of all rules that are
        valid at the time represented by the given datetime object,
        keeping the order from the rules list. SubScheduleRule objects
        are expanded and their matching rules are included."""

        _time = when.time()
        for path in self.unfold():
            for path_idx, rule in enumerate(path):
                break_path = False
                days_back = -1
                found_start_day = False
                while days_back < rule.end_plus_days:
                    days_back += 1
                    # starts with days=0 (meaning the current date)
                    _date = when.date() - datetime.timedelta(days=days_back)

                    found_start_day = found_start_day or \
                                      rule.check_constraints(_date)
                    if not found_start_day:
                        # try next day
                        continue

                    # in first loop run, rule has to start today and not
                    # later than now (rule start <= when.time())
                    if days_back == 0 and rule.start_time > _time:
                        # maybe there is a next day to try out
                        continue

                    # in last loop run, rule is going to end today and that
                    # has to be later than now (rule end > when.time())
                    if days_back == rule.end_plus_days and \
                       rule.end_time <= _time:
                        # rule finally disqualified, continue with next path
                        break_path = True
                        break

                    # rule matches!
                    if isinstance(rule, Rule):
                        yield path[:path_idx + 1]
                    break

                if break_path:
                    # continue with next path
                    break

    def next_schedule_datetime(
            self, now: datetime.datetime
    ) -> T.Optional[datetime.datetime]:
        """Returns a datetime object with the time at which the next
        re-scheduling should be done. now should be a datetime object
        containing the current date and time.
        SubScheduleRule objects and their rules are considered as well.
        None is returned in case there are no rules in the schedule
        which are not universally valid anyway."""

        times = set()  # type: T.Set[datetime.time]
        for path in self.unfold():
            for rule in path:
                if not rule.is_always_valid():
                    times.update((rule.start_time, rule.end_time),)
        if not times:
            # no constrained rules in schedule
            return None

        current_time = now.time()
        today = now.date()
        tomorrow = today + datetime.timedelta(days=1)

        def map_func(_time: datetime.time) -> datetime.datetime:
            """Maps a time object to a datetime containing the next
            occurrence of that time. Midnight transitions are handled
            correctly."""

            if _time < current_time:
                # midnight transition
                return datetime.datetime.combine(tomorrow, _time)

            return datetime.datetime.combine(today, _time)

        return min(map(map_func, times))

    def unfold(self) -> T.Iterator[RULE_PATH_TYPE]:
        """Returns an iterator over rule paths (tuples of Rule objects).
        The last element of each tuple is a Rule object, the elements
        before - if any - represent the chain of SubScheduleRule objects
        that led to the final Rule."""

        for rule in self.rules:
            if isinstance(rule, SubScheduleRule):
                _rule = rule  # type: Rule
                for path in rule.sub_schedule.unfold():
                    yield (_rule,) + path
            else:
                yield (rule,)


def get_rule_path_temp(path: RULE_PATH_TYPE) -> Rule:
    """Returns the first rule containing a temperature expression,
    searching the path from right to left. A ValueError is raised in
    case there is no rule with a temperature expression in the path."""

    for rule in reversed(path):
        if rule.temp_expr is not None:
            return rule

    raise ValueError("No temperature specified for any rule along the path: {}"
                     .format(path))
